!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!                                                              
!     Apparent Horizon Finder for Schwarzschild BH             
!                                  in isotropic coordinates    
!                                                              
!--------------------------------------------------------------
!                                                              
!     Sample Code for Lecture in NR/HEP2 Spring School         
!                                                              
!                                     Coded by Hirotada Okawa  
!                                                              
!==============================================================
!   How to compile and use this program in terminal(bash)
!==============================================================
!   $ gfortran -O2 -ffast-math -o ahf_sbh sor_AHF_SBH_ISO.f90
!   $ ./ahf_sbh
!--------------------------------------------------------------
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

module inc_coord
  implicit none

!--------------------------------------------------------------
!     Grid points
!--------------------------------------------------------------
  integer,parameter :: jli=1
  integer,parameter :: jui=60
  integer,parameter :: jlb=jli-1
  integer,parameter :: jub=jui+1

!--------------------------------------------------------------
!     Maximum/Minimum Coordinates (physical)
!--------------------------------------------------------------
  real(8),parameter :: pi=atan(1.d0)*4.d0
  real(8),parameter :: xlower=0.
  real(8),parameter :: xupper=pi
  real(8),parameter :: dx=(xupper-xlower)/dble(jub-jli)
  real(8),parameter :: dxi=1.d0/dx

!--------------------------------------------------------------
!     Array for Coordinates (physical)
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: x

!--------------------------------------------------------------
!     Variables to solve
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: h, hprev

!--------------------------------------------------------------
!     Source term for Poisson equation
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: src
end module inc_coord

program main
  use inc_coord
  implicit none

!--------------------------------------------------------------
!     Definition of parameters
!--------------------------------------------------------------
  integer,parameter :: stepmax=1d8
  real(8),parameter :: errormax=1.d-8
  real(8),parameter :: fpar=1.d0
  real(8),parameter :: omega=1.d0

!--------------------------------------------------------------
!     Definition of temporary variables to use
!--------------------------------------------------------------
  integer :: j, step
  real(8) :: xx, hh, hx, hxx, hhi, h1
  real(8) :: errortmp, vtmp
  real(8) :: cosx, sinx

!--------------------------------------------------------------
!     Output File
!--------------------------------------------------------------
  open(200,file='data_sbh.dat')

!--------------------------------------------------------------
!     Initialization
!--------------------------------------------------------------
  do j=jlb,jub
     x(j)  = xlower +(dble(j)-0.5d0)*dx    ! Coordinates
     h(j)  = dcos(xx)                      ! variable to solve
     hprev(j) = h(j)                       ! previous variable
  end do

!**************************************************************
!     Main Loop
!**************************************************************
  do step=0,stepmax

!--------------------------------------------------------------
!     Preserve data of previous step
!--------------------------------------------------------------
     do j=jlb,jub
        hprev(j) = h(j)
     end do

!--------------------------------------------------------------
!     SOR Method
!--------------------------------------------------------------
     do j=jli,jui

!==============================================================
!     Definition of Source term
!==============================================================
        xx  = x(j)
        sinx= dsin(xx)
        cosx= dcos(xx)
        hh  = h(j)
        hhi = 1.d0/hh
        h1  = 1.d0 +hhi**2
        hx  = (h(j+1) -h(j-1))*0.5d0*dxi
        hxx = (h(j+1) +h(j-1) -2.d0*hh)*dxi**2

        src(j) = -2.d0/(1.d0+0.5d0*hhi) +2.d0*hh &
&                -(1.d0+hhi**2*hx**2)*cosx/sinx*hx &
&                -2.d0*hx**2/(hh**2 +hh*0.5d0) &
&                -hx**2*hhi**2*hxx

        h(j) = (1.-omega)*h(j)+omega*0.5d0*( hprev(j+1) +h(j-1) -dx**2*src(j) )
     end do

!==============================================================
!     Impose Boundary Condition
!==============================================================
     h(jub)=h(jui)         ! Neumann Boundary Condition
     h(jlb)=h(jli)         ! Neumann Boundary Condition

!--------------------------------------------------------------
!     Check if values converge
!--------------------------------------------------------------
     errortmp=0.d0
     vtmp=0.d0
     do j=jli,jui
        errortmp = errortmp +(h(j)-hprev(j))*(h(j)-hprev(j))*dx**2
        vtmp = vtmp + dx**2
     end do
     errortmp = dsqrt(errortmp/vtmp)
     if( (errortmp.le.errormax) .and. (step.gt.1) ) exit

!--------------------------------------------------------------
!     Next Guess(Mix Previous value) just in case
!--------------------------------------------------------------
     do j=jlb,jub
        h(j) = fpar*h(j) +(1.d0-fpar)*hprev(j)
     end do

     write(*,*) "Step=",step,"Error=",errortmp
  end do

!--------------------------------------------------------------
!     Print Data
!--------------------------------------------------------------
  do j=jlb,jub
     write(200,'(4e16.8e2)') x(j),h(j),hprev(j),src(j)
  end do

!--------------------------------------------------------------
!     End of Program
!--------------------------------------------------------------
  write(*,*) "End of Run",errortmp
  close(200)

end program main
